--[[
Generate Preset-Groups - v1.0.0
Created by Jason Giaffo
Contact: http://GiaffoDesigns.com/contact/
Last updated Jan 26, 2018

-- Creates groups based on a range of presets, to easily select all fixtures used in a given preset.
-- An additional filter based on sequence ranges is provided.
--]]



local config = {}

config.grp_suffix = ' FIXT'                       -- text attached to preset name
config.include_fixt_count = true                  -- possible values: true, false














local GDFstr = [===[
-- GD Functions Module
-- Version 1.0.6

-- Include this plugin before any other plugins in your plugin pool
-- (i.e #2 if possible)
-- The functions in this plugin run any plugin including the line
-- "GDF = GDF_1_0_X" up to 1_0_6 minumum


local _M = {}
local unpack = table.unpack

--[=[ Require and shortcut chunk
local GDF = require [[GD/_GDF_v0-0-3a]]

local class    = GDF.class
local lo    = GDF.lo
local gui    = GDF.gui
local dmx    = GDF.dmx
local env    = GDF.env
local data    = GDF.data
local tbl    = GDF.table
local xml    = GDF.xml
local csv    = GDF.csv
local lua    = GDF.lua
local get    = GDF.get
local set    = GDF.set
local pi    = GDF.poolitem
local plugin= GDF.plugin
local exec    = GDF.exec

local self_table = {...}
--]=]

---- "Object" Classes

_M.class = { --sub-module v1.1.0
    superTable = {
        append = function(self, v)
            self[#self + 1] = v
        end,

        concat = function(self, sep, i, j)
            return table.concat(self, sep, i, j)
        end,
        
        new = function(self, o)
            o = o or {}
            setmetatable(o, self)
            self.__index = self
            return o
        end,

        shuffle = function(list)
            local iterations = #list
            for x = 1, 4 do                                --repeat process, else last number ends up the same
                for i = iterations, 1, -1 do
                    j = math.random(i)
                    list[i], list[j] = list[j], list[i]
                end
            end
        end
    }
}

local ST = _M.class.superTable

_M.class.macro = _M.class.superTable:new{        -- obj v2.0.0
    -- create new macro with line:
    -- mymacro = class.macro:new{num = num, name = name}
    
    import = function(self, ref)
        if not _M.get.verify('Macro '..ref) then return nil, 'Error: requested macro does not exist' end
        
        local script = _M.get.script('Macro '..ref, 'macros')
    
        self.name = _M.get.name('Macro '..ref)
        self.num = _M.get.number('Macro '..ref)

        for match in script:gmatch('<Macroline index=.-</Macroline>') do
            local t = {}
            
            local num = match:match('Macroline index="(%d+)"')    -- extract macro number
            t.num = tonumber(num) + 1
            if not self.lines then self.lines = ST:new() end
            while t.num > #self.lines do                        -- offset for empty lines
                self:newline({})
            end
            
            local cmd = match:match('<text>(.-)</text>')        -- extract command
            if cmd then t.cmd = cmd:gsub('&quot;', '"') end
            
            
            local info = match:match('<info>(.-)</info>')        -- extract info
            if info then t.info = info:gsub('&quot;', '"') end
            
            local wait = match:match('delay = \"(.-)\"')        -- extract wait time
            if wait then t.wait = tonumber(wait) end    
        
            self:newline(t)                                    -- add line to macro array
        end
    end,
    
    newline = function(self, list)                -- create function for appending new lines to macro
        if not self.lines then self.lines = _M.class.superTable:new() end
        self.lines:append{                                        -- list keys: cmd, wait, info, disabled
            cmd = list.cmd,                -- type: str
            wait = list.wait,            -- type: num
            info = list.info,            -- type: str
            disabled = list.disabled    -- type: bool
}
    end,

    write = function(self, force_overwrite)        -- function to write macro to showfile
        local num = self.num
        local name = self.name or 'Macro '..tonumber(num)
        local script = [[
<?xml version="1.0" encoding="utf-8"?>
<MA xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns="http://schemas.malighting.de/grandma2/xml/MA" xsi:schemaLocation="http://schemas.malighting.de/grandma2/xml/MA http://schemas.malighting.de/grandma2/xml/3.2.2/MA.xsd" major_vers="3" minor_vers="2" stream_vers="2">
    <Info datetime="2017-01-07T23:39:00" showfile="Ye Olde Bag O Phallus" />
    <Macro index="1" name="]]..name..[[">"]]..'\n'
    
        
        for i, v in ipairs(self.lines) do
            local cmd    = self.lines[i].cmd
            local wait    = self.lines[i].wait
            local info    = self.lines[i].info
            local disabled    = self.lines[i].disabled
            
          if cmd  then cmd    = cmd:gsub([["]], [[&quot;]]) end
          if info then info = info:gsub([["]], [[&quot;]]) end

          --build XML script to append to master script
          local t = [[
        <Macroline index="]]..tostring(i-1)..[["]]
          
          if type(wait) == 'string' and wait:upper() == 'GO' then
            t = t..[[ delay="-1" ]]
          elseif wait then
            t = t..[[ delay="]]..wait..[["]] end
            
          if disabled then t = t..[[ disabled="true"]] end
          
          --command and info lines
          if (not cmd) and (not info) then t = t..[[ />]]..'\n'
          else
            t = t..[[>]]..'\n'
            
            if cmd then t  = t..[[
            <text>]]..cmd..[[</text>]]..'\n' end
            
            if info then t = t..[[
            <info>]]..info..[[</info>]]..'\n' end
            
            
            --close out script for macro line
            t = t..[[
                </Macroline>]]..'\n'
                
            --append new XML text to master script
            script = script..t
          end
        end
        
        --close out script for entire macro
          script = script..[[
    </Macro>
</MA>]]
        
        local macFile = {}
        macFile.name        = 'tempFile_createMacro.xml'
        macFile.dir            = gma.show.getvar('PATH')..'/'..'macros/'
        macFile.fullpath    = macFile.dir..macFile.name  

        local writeFile = io.open(macFile.fullpath, 'w')
        writeFile:write(script)
        writeFile:close()

        local importcmd = 'Import \"'..macFile.name..'\" Macro '..num        -- build string for import
        if force_overwrite then importcmd = importcmd..' /o /nc' end        -- (if force-overwrite enabled)
        gma.cmd('SelectDrive 1')    
        gma.cmd(importcmd)
        os.remove(macFile.fullpath)

        gma.cmd('Label Macro '..num..' \"'..name..'\"')
    end,
    
    
}


    
---- Layouts
_M.lo = {
    layout = ST:new{
        name = nil,    -- string value: name of layout when imported
        num = nil,    -- number value: number where layout will be imported
        
        size  = {x = 0, y = 0},
        bound = {x = {}, y = {}},
        --\\ format: bound.(x/y).(min/max)
        
        rebound = function(self, obj)
            if not obj.bound then obj.bound = {
                x = {
                    min = obj.coord.x - (obj.size.x/2),
                    max = obj.coord.x + (obj.size.x/2),
                },
                y = {
                    min = obj.coord.y - (obj.size.y/2),
                    max = obj.coord.y + (obj.size.y/2),
                }
            }
            end
            
            -- adjust master layout bound values
            if not self.bound.x.min then            -- create bounds for layout if first object
                self.bound = obj.bound
            else                                    -- modify bounds for new objects added
                for k, v in pairs(self.bound) do
                    -- set maximums
                    self.bound[k].max = math.max(v.max, obj.bound[k].max)

                    -- set minimums
                    self.bound[k].min = math.min(v.min, obj.bound[k].min)
                end
            end
        
            -- reset size values
            for k, v in pairs(self.bound) do
                self.size[k] = math.abs(self.bound[k].max - self.bound[k].min)
            end
            
            -- gma.echo('CURRENT LAYOUT BOUNDS:')                                                    -- TROUBLESHOOT
            -- gma.echo(self.bound.x.min, ' ', self.bound.x.max)
            -- gma.echo(self.bound.y.min, ' ', self.bound.y.max)
            -- gma.echo('')
        end,
        
        -- Layout Object Types
        obj = ST:new{        
            id = 'obj',                -- object default. Do not change.
            
            obj = {
                type = nil,            -- string. Examples 'plugin', 'macro', 'group'
                num    = nil            -- number. Example: for [Macro 501], enter 501 here.
            },
            
            img = nil,                -- number.
            
            coord = {
                x = 0,
                y = 0,
            },
            
            size = {
                x = 0.95,
                y = 0.95
            },

            icon = {
                vis = 'Simple',        -- valid options: 'pool icon', 'simple' (groups, pool items), filled, spot (groups, fixtures). Use 'simple' for images
                bg = '3c3c3c',        
                border = '5a5a5a',    -- rrggbb hex string
                icon = 'None',        -- enter icon number if icon is being used
            },
            
            text = {
                text = '',            -- string
                size = 2,            -- options: 1, 2, 3. Does not apply to text in "text" field above, but to name only.
                show_id = false,    -- bool
                show_name = false,    -- bool
                show_type = false,    -- bool
            }
            
            
        },
            
        box = ST:new{
            id = 'box',
            
            bg = '00000000',
            border = 'ffffff',
            
            img = nil,    -- number value
            
            coord = {
                x = 0,
                y = 0,
            },
            
            size = {
                x = 0.95,
                y = 0.95,
            },
            
            text = {
                text = '',
                show_id = true,
                show_name = true,
                show_type = true,
            },
            
            active = {
                show_dVal = false,
                show_dBar = false,
                select_grp = false,
            },
        
        
        },


        -- collection of all generated objects
        collection = ST:new{
            obj = ST:new(),
            box = ST:new(),
        },
            
        
        -- function for adding new object to layout
        addobj = function(self, config)
            local l = _M.lo.layout    -- local "directory"
            
            local t = l.obj:new(config)
            l.collection.obj:append(t)
            self:rebound(t)
            return t
        end,
        
        addbox = function(self, config)
            local l = _M.lo.layout    -- local "directory"
            
            -- if box entered as bounds, convert to coord+size values
            if config.bounds then
                local d = config.bounds                -- "local directory"
                config.size = {
                    x = math.abs(d.x[2] - d.x[1]),
                    y = math.abs(d.y[2] - d.y[1])
                }
                
                config.coord = {
                    x = (d.x[1] + d.x[2])/2,
                    y = (d.y[1] + d.y[2])/2,
                }
            end
            
            local t = l.box:new(config)
            l.collection.box:append(t)
            self:rebound(t)
            return t
        end,
        
        -- layout transform methods
        shift = function(self, x, y)
            if x and y and type(x) == 'number' and type(y) == 'number' then
                -- translate objects inside collection
                for _, v in pairs(self.collection) do
                    for _, obj in ipairs(v) do
                        obj.coord.x = obj.coord.x + x
                        obj.coord.y = obj.coord.y + y
                    end
                end
                
                -- translate master x boundaries
                for k, v in pairs(self.bound.x) do
                    self.bound.x[k] = v + x
                end
                
                -- translate master y boundaries
                for k, v in pairs(self.bound.y) do
                    self.bound.y[k] = v + y
                end
            else
                return nil, 'error: invalid data for LO-shift'
            end
        end,
        
        scale = function(self, x, y)
            local scl = {
                x = x or 1,
                y = y or 1,
            }
            
            local avg = function(a, b)
                if type(a) == 'table' and (not b) then
                    return ((a.min + a.max)/2)
                elseif type(a) == 'table' and type(b) == 'table' then
                    return ((a + b)/2)
                end
            end
            
            ---- Reset master boundaries of object ----
            -- master center info
            local mast = {
                cent = {
                    x = avg(self.bound.x),
                    y = avg(self.bound.y),
                },
                
                size = {
                    x = (self.bound.x.max - self.bound.x.min) * scl.x,
                    y = (self.bound.y.max - self.bound.y.min) * scl.y
                }
            }
            
            for k, v in self.bound do
                self.bound[k].min = mast.cent[k] - (mast.size[k]/2)
                self.bound[k].max = mast.cent[k] + (mast.size[k]/2)
            end
            
            ---- Reset Individual items position + size ----
            for type, v in pairs(self.collection) do
                for i, obj in ipairs(v) do
                    local l = self.collection[type][i]    -- local "directory" = layout object
                        l.coord.x = l.coord.x + ((l.coord.x - mast.cent.x) * scl.x)
                        l.coord.y = l.coord.y + ((l.coord.y - mast.cent.y) * scl.y)
                        
                        l.size.x = l.size.x * scl.x
                        l.size.y = l.size.y * scl.y
                end
            end
        end,
        
        merge = function(self, other)
        self:rebound(other)                -- merge boundaries
        
        for type, table in pairs(other.collection) do    -- merge other's object tables into self
            for i, v in ipairs(table) do
                self.collection[type]:append(v)
            end
        end
    end,
    
        -- write, import into showfile
        write = function(self, config)
            local config = config or {}
            
            local trunc = _M.data.trunc
        
            local scr = ST:new()
            -- still need
            --\\ name
            --\\ number
        
        
            local function write_obj(scr, obj_table)
                -- header for "objects" section
                scr:append([[
            <CObjects>
]])

                -- body for "objects"
                for _, v in ipairs(obj_table) do
                    ---- Process object information ----
                    local x = trunc(v.coord.x, 0.001)
                    local y = trunc(v.coord.y, 0.001)
                    local size_x = trunc(v.size.x, 0.001)
                    local size_y = trunc(v.size.y, 0.001)
                    
        
                    local show_id, show_name, show_type
                    if v.text.show_id == true then show_id = [[show_id="1" ]]; else show_id = ''; end;
                    if v.text.show_name == true then show_name = [[show_name="1" ]]; else show_name = ''; end;
                    if v.text.show_type == true then show_type = [[show_type="1" ]]; else show_type = ''; end;

                
                
                    ---- Object header ----
                    scr:append([[
                <LayoutCObject font_size="Small" center_x="]]..x..[[" center_y="]]..y..[[" size_h="]]..size_y..[[" size_w="]]..size_x..[[
" background_color="]]..v.icon.bg..[[" border_color="]]..v.icon.border..[[" icon="]]..v.icon.icon..[[" ]]..show_id..show_name..show_type..[[function_type="]]..v.icon.vis..[[" select_group="1">
]]);
                
                
                
                    ---- Image Information ----
                    --if v.img.name and v.img.rootnum then    -- if valid entry provided for image information
                    if tonumber(v.img) and tonumber(v.img) > 0 and _M.get.verify('Image '..v.img) then
                        local rootnum = _M.get.rootnum('Image', v.img)
                        local name = _M.get.label('Image '..v.img)
                        scr:append([[
                    <image name="]]..name..[[">
                        <No>]]..rootnum[1]..[[</No>
                        <No>]]..rootnum[2]..[[</No>
                    </image>
]]);
                    else                                    -- append string for "no image"
                        scr:append([[
                    <image />
]])
                    end
                
                
                    ---- Object Information ----
                    -- head/name
                    for k, v in pairs(v.obj) do                                                                        -- TROUBLESHOOT
                        gma.echo(k, ' - ', type(v), ' - ', v)
                    end
                    
                    local w_obj = {                                    -- create write-object data
                        name = _M.get.label(v.obj.type..' '..v.obj.num),
                        rootnum = _M.get.rootnum(v.obj.type, v.obj.num)
                    }

                    scr:append([[
                    <CObject name="]]..w_obj.name..[[">
]]);

                    -- object root number
                    for i = 1, #w_obj.rootnum do
                        scr:append([[
                        <No>]]..w_obj.rootnum[i]..[[</No>
]]);
                    end
            
                    -- close
                    scr:append([[
                    </CObject>
]]);                
                
                    ---- Object Footer ----
                    scr:append([[
                </LayoutCObject>
]]);
                end
            
            
                -- footer for "objects" section
                scr:append([[
            </CObjects>
]])
            end
            
            local function write_box(scr, box_table)
                ---- Rectangles Header ----
                scr:append([[
            <Rectangles>
]]);

                ---- Rectangles Body ----
                for _, v in ipairs(box_table) do
                    ---- Process object information ----
                    local x = trunc(v.coord.x, 0.001)
                    local y = trunc(v.coord.y, 0.001)
                    local size_x = trunc(v.size.x, 0.001)
                    local size_y = trunc(v.size.y, 0.001)

                    local show_id, show_name, show_type, show_dVal, show_dBar
                    if v.text.show_id == true then show_id = [[show_id="1" ]]; else show_id = ''; end;
                    if v.text.show_name == true then show_name = [[show_name="1" ]]; else show_name = ''; end;
                    if v.text.show_type == true then show_type = [[show_type="1" ]]; else show_type = ''; end;

                    local show_dimmerVal; if v.active.show_dVal then show_dimmerVal = [[show_dimmer_value="Off" ]]; else show_dimmerVal = ''; end;
                    local show_dimmerBar; if v.active.show_dBar then show_dimmerBar = [[show_dimmer_bar="Off" ]]; else show_dimmerBar = ''; end;
                    local select_grp; if v.active.select_grp then select_grp = [[select_group="1" ]]; else select_grp = ''; end;
                    
                    -- start rectangle object
                    scr:append([[
                <LayoutElement font_size="Small" center_x="]]..x..[[" center_y="]]..y..[[" size_h="]]..size_y..[[" size_w="]]..size_x..[[
" background_color="]]..v.bg..[[" border_color="]]..v.border..[[" icon="None" text="]]..v.text.text..[[" ]]..show_id..show_name..show_type..show_dimmerBar..show_dimmerVal..select_grp..[[>
]]);

                    -- Image Information
                    if tonumber(v.img) and tonumber(v.img) > 0 and _M.get.verify('Image '..v.img) then        -- if image info provided
                        local rootnum = _M.get.rootnum('Image', v.img)
                        local name = _M.get.label('Image '..v.img)
                        scr:append([[
                    <image name="]]..name..[[">
                        <No>]]..rootnum[1]..[[</No>
                        <No>]]..rootnum[2]..[[</No>
                    </image>
]]);
                    else                                        -- if no image
                        scr:append([[
                    <image />
]]);
                    end
                
                    -- end rectangle object
                    scr:append([[
                </LayoutElement>
]])

                end
            
                ---- Rectangles Footer ----
                scr:append([[
            </Rectangles>
]])                
            end




            -- write file header
            scr:append([[
<?xml version="1.0" encoding="utf-8"?>
<MA xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns="http://schemas.malighting.de/grandma2/xml/MA" xsi:schemaLocation="http://schemas.malighting.de/grandma2/xml/MA http://schemas.malighting.de/grandma2/xml/3.2.2/MA.xsd" major_vers="3" minor_vers="1" stream_vers="2">
    <Info datetime="2016-10-18T14:28:09" showfile="xml_scripting" />
    <Group index="0" name="]]..self.name..[[">
        <LayoutData index="0" marker_visible="true" background_color="000000" visible_grid_h="1" visible_grid_w="1" snap_grid_h="0.5" snap_grid_w="0.5" default_gauge="Filled &amp; Symbol" subfixture_view_mode="DMX Layer">
]]);

            -- write layout objects
            if #self.collection.obj > 0 then
                write_obj(scr, self.collection.obj)
            end
        
            -- write boxes
            if #self.collection.box > 0 then
                write_box(scr, self.collection.box)
            end
        
                
            ---- write file footer
            scr:append([[
        </LayoutData>
    </Group>
</MA>]]);

            ---- Filesystem Info ----
            local fn = {temp = 'temp_layoutgen.xml'}
            local dir = {}
                dir.main = gma.show.getvar('PATH')..'/'
                dir.imex = dir.main..'importexport/'
            local file = {temp = dir.imex..fn.temp}
            
            local wf = io.open(file.temp, 'w')
            wf:write(scr:concat())                    -- concatenate script lines and write
            wf:close()
        
            -- Import Layout to designated slot
            gma.cmd('Import \"'..fn.temp..'\" Layout '..self.num)
            
            -- remove temp file
            if not config.debug then
                os.remove(file.temp)
            end
            
        end    --end of write method

    },
    
    spacingTable = function(tbl)
        --[==[\\ input: {
            dim = {cols, rows}, --(number of objects on x-axis, on y-axis)
            space = {x, y},    -- distance between BOUNDARIES of objects
            size = {x, y},    -- size of objects
            start = {x, y},    -- reference point for generated coordinates
            iscent = {true, true},        -- set to true for POS entry to reference center of generated grid, instead of top/left
            first = {x = true},        -- set to {y = true} to traverse y-axis first
        }
        ]==]
        --\\ output: spacingtable[i] = {xvalue, yvalue}
        --assimilate data
        local x = {
            ct        = tbl.dim[1]                        or 1,
            int        = tbl.space[1] + tbl.size[1]        or 1,
        }
        if tbl.iscent[1] then x.start = tbl.start[1] - ((x.ct-1) * x.int * 0.5) end
        x.pos = x.start
        
        local y = {
            ct        = tbl.dim[2]                        or 1,
            int        = tbl.space[2] + tbl.size[2]        or 1,
        }
        if tbl.iscent[2] then y.start = tbl.start[2] - ((y.ct-1) * y.int * 0.5) end
        y.pos = y.start


        -- create output table
        local t = ST:new{}
        local loops = 1
        
        -- there's a better way to do this if I had more coffee
        if tbl.first.y then                        ---- process for y-first
            for col = 1, x.ct do
                for row = 1, y.ct do
                    t[loops] = {x.pos, y.pos}    -- add position to table
                    y.pos = y.pos + y.int        -- advance y position
                    loops = loops + 1            -- advance loop count
                end
                
                x.pos = x.pos + x.int            -- advance column position
                y.pos = y.start                    -- reset row position
            end    
        else                                    ---- process for x-first (default)
            for row = 1, y.ct do
                for col = 1, x.ct do
                    t[loops] = {x.pos, y.pos}    -- add position to table
                    x.pos = x.pos + x.int        -- advance x position
                    loops = loops + 1            -- advance loop count
                end
                
                y.pos = y.pos + y.int            -- advance row position
                x.pos = x.start                    -- reset column position
            end
        end
        
        
        t.index = 0
        t.next = function(self)
            self.index = self.index + 1
            local i = self.index
            if self.index <= #self then
                return {x = self[i][1], y = self[i][2]}
            else
                return nil, 'error: final index passed'
            end
        end
        return t        
    end,

}


    
---- UI Functions

_M.gui = {    --sub-module version 1.1.0
-- cmdview() requires advanceSpace function
    superBar = {
        value = 0,

        new = function(self, name)
            o = {handle = gma.gui.progress.start(name)}

            setmetatable(o, self)
            self.__index = self
            return o
        end,
        
        set = function(self, num, add)
            if add then num = self.value + num end
            
            num = math.floor(num)
            gma.gui.progress.set(self.handle, num)
            self.value = num
        end,
        
        setrange = function(self, bottom, top)
            if not bottom or not top then return nil end
            gma.gui.progress.setrange(self.handle, bottom, top)
            self.range = {top = top, bottom = bottom}
        end,
        
        settext = function(self, text)
            gma.gui.progress.settext(self.handle, text)
        end,
        
        rename = function(self, name)
            gma.gui.progress.stop(self.handle)
            self.handle = gma.gui.progress.start(name)
            if self.range then gma.gui.progress.setrange(self.handle, self.range.bottom, self.range.top) end
        end,
        
        stop = function(self)
            gma.gui.progress.stop(self.handle)
        end,
        
        time_move = function(self, target, time, add, ignore_top)
            if add then target = self.value + target end
            local start = self.value
            local finish = math.floor(target)
            
            local sleepPeriod
            if math.abs(time) ~= time then sleepPeriod = (1 / time)                                            --allow for negative time values to trigger a "rate" function
            else sleepPeriod = time / (math.abs(finish - start)) end
            
            if not ignore_top and self.range and finish > self.range.top then finish = self.range.top end    --set finish point to top of progress-bar range unless specified to ignore
            
            local dir = 1
            if start > finish then dir = -1 end
            
            for i = start, finish, dir do
                self:set(i)
                gma.sleep(sleepPeriod)
            end    
        end
    }, --end of superBar object
    
    
    msgbox = function(title, message_box, message_text)
      -- v. 1.0
      -- function avoids using confirmation box function with version 3.1.2.5, where it doesn't exist
      local confirm_method
      
      local version = gma.show.getvar('version')
      
      if version:find('3.1.2') == 1 then 
        confirm_method = 'textinput'
      else 
        confirm_method = 'box' end
      
      if confirm_method == 'box' then
        gma.gui.msgbox(title, message_box)
      elseif confirm_method == 'textinput' then
        gma.textinput(title, message_text)
      end
    end,
    
    
    confirm = function(title, message_box, message_text)
      -- v. 1.0
      -- function avoids using confirmation box function with version 3.1.2.5, which crashes the software
      local confirm_method
      
      local version = gma.show.getvar('version')
      
      if version:find('3.1.2') == 1 then 
        confirm_method = 'textinput'
      else 
        confirm_method = 'box' end
      
      if confirm_method == 'box' then
        return gma.gui.confirm(title, message_box)
      elseif confirm_method == 'textinput' then
        local t = gma.textinput(title, message_text)
        if t then t = true end
        return t
      end
    end,
    
    
    print = function(str)
        str = tostring(str)
        for line in str:gmatch('[^\n]+') do
            gma.echo(line)
            gma.feedback(line)
        end
    end,
    
    
    windowSize = function(ct, maxwidth)
        local t = {}

        if ct <= (maxwidth - 1) then
            t.y = 1
            t.x = ct
            return t
        else
            local testnum = math.ceil(ct/maxwidth)
            if ((maxwidth * testnum) - 1) < ct then t.y = testnum + 1
            else t.y = testnum end


            local capacity = (maxwidth * t.y - 1)
            local excess = capacity - ct

            t.x = maxwidth - math.floor(excess/t.y)

            return t
        end
    end,
    
    
    cmdview = function(screen)
        if not screen then screen = 2 end
        local cmd_script = [[
<?xml version="1.0" encoding="utf-8"?>
<MA xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns="http://schemas.malighting.de/grandma2/xml/MA" xsi:schemaLocation="http://schemas.malighting.de/grandma2/xml/MA http://schemas.malighting.de/grandma2/xml/3.2.2/MA.xsd" major_vers="3" minor_vers="2" stream_vers="2">
    <Info datetime="2017-06-06T19:11:55" showfile="some shit" />
    <View index="899" name="CMD/Sys" display_mask="2">
        <BitMap width="96" height="48">
            <Image>
                dBUY/09RYP95Cgz/gAIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4MGBv+BAwP/
                gQIC/4ECAv+BAgL/gQIC/4EDA/+DBwf/gQIC/4ECAv+CBAT/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                gQIC/4ECAv+AAgL/gAIC/4ACAv9/BAT/RlRk/wSryf9NKDD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/
                fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/
                fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/
                fwAA/38AAP9/AAD/fwAA/6tYWP+uXl7/pk9P/6lVVf+sWlr/tWtr/5gyMv/Fi4v/rFpa/6lUVP+1a2v/
                qFNT/6FFRf9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/
                fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/
                fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP+BAgL/cQ4Q/0YnLv92Bwf/
                gAIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4MHB/+MGBj/gwYG/4IFBf+DBwf/
                hAgI/4IEBP+DBgb/gwYG/4MGBv+DBgb/hAgI/4IEBP+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+AAgL/
                gAIC/4ACAv9/BAT/Fj1M/yA7Rf8kP0n/IjxF/x03Qf8fOUL/GjQ+/xo0Pv8fOUL/CiQt/wslL/8KMzz/
                Ci02/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/
                CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/
                CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/
                CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/
                CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/
                CiQt/wokLf8KIy3/CSQt/wkkLv8OMT7/CCo2/wciLP8LM0H/HTdA/zAwMP8zMzP/Kysr/zAwMP8oKCj/
                JSUl/ygoKP8rKyv/AAAA/wYGBv8DGxv/AA0N/wYGBv8BAQH/AAAA/woKCv8HBwf/AAAA/wEBAf8AAAD/
                DAwM/wQEBP8BAQH/AwMD/wAAAP8HBwf/AAAA/wAAAP8AAAD/AAAA/wEBAf8AAAD/BAQE/wAAAP8AAAD/
                AAAA/wEBAf8GBgb/AwMD/w0NDf8PDw//AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IISv/ACMx/wFObP8GIy7/
                GjQ9/y0tLf82Njb/JSUl/zExMf8lJSX/IiIi/yUlJf8wMDD/AAAA/wYGBv8eHh7/KCgo/y0tLf8lJSX/
                JCQk/xYWFv8ZGRn/MDAw/yIiIv8KCgr/IiIl/w8PD/8iIiT/JSUl/xgYGP8WFhb/MzM3/xAQEP80NDb/
                MDAw/ygoKP8cHBz/HBwf/yQkJP8qKir/BgYG/yUlJf8PDw//EhIS/yoqKv8tLS3/MDAw/wkJCf8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAB/wAAAf8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAD/wAAAP8AAAD/AAAB/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8IIiv/ACAs/wATG/8GJzP/HjhC/zMzM/85OTn/MDAw/zY2Nv8kJCT/JSUl/y0tLf8rKyv/
                AAAA/wYGBv8AAC3/AAAk/wAAJP8AACX/AAAW/wAAJP8AAB//AAAc/wAABv8AAAD/AAAi/wAAH/8AACT/
                AAAh/wAAJP8AABj/AAAu/wAAIf8AADf/AAAc/wAAHv8AACf/AAAZ/wAADP8AADP/AAAH/wAAFv8AACj/
                AAAk/wAAKP8AABP/AAAl/wAAJf8AAAT/AAAD/wAABv8AAAT/AAAQ/wAAIf8AAC3/AAAf/wAAOv8AAAr/
                AAAW/wAAG/8AACj/AAAn/wAANv8AAAP/AAAo/wAAKP8AAB//AAAh/wAAJf8AAA3/AAAn/wAAKv8AAAT/
                AAAH/wAAA/8AAAf/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/Fi43/xglKf8VNUH/HjhC/x8fH/8xMTH/
                HBwc/xYWFv8ZGRn/LS0t/y4uLv8rKyv/Jycn/xAQEP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/
                GCUp/xoaGv8WMDn/HjhC/zQ0NP86Ojr/MDAw/zY2Nv8kJCT/JSUl/y0tLf8rKyv/AAAA/wYGBv8fHx//
                KCgo/zExMf8nJyf/HBwc/x4eHv8fHx//KCgo/yQkJP8KCgr/Kysr/xISEv8kJCT/Jycn/xgYGP8cHBz/
                MzMz/xAQEP80NDT/KCgo/yoqKv8cHBz/Hx8f/yQkJP8qKir/BgYG/ycnJ/8AAAD/AAAA/y4uLv8kJCT/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/GCUp/xoaGv8WMDn/HTdA/zAwMP8uLi7/MDAw/zAwMP8oKCj/
                IiIi/yUlJf8nJyf/AAAA/wQEBP8cHBz/JCQk/y0tLf8iIiL/Hx8f/xwcHP8cHBz/Kioq/yEhIf8JCQn/
                IiIi/w0NDf8hISH/IiIi/xUVFf8ZGRn/LS0t/w0NDf8uLi7/Kioq/yUlJf8ZGRn/HBwc/x8fH/8lJSX/
                BgYG/yIiIv8AAAD/AAAA/yoqKv8fHx//AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/GCUp/xoaGv8WMDn/
                HTdA/zAwMP82Njb/Nzc3/zExMf8qKir/JCQk/ycnJ/8lJSX/AAAA/wYGBv8nJyf/KCgo/xkZGf8eHh7/
                Li4u/ycnJ/8WFhb/NjY2/y4uLv8cHBz/FhYW/wQEBP8tLS3/IiIi/xgYGP8iIiL/LS0t/xkZGf8nJyf/
                Hx8f/ygoKP8NDQ3/DAwM/w8PD/8NDQ3/AwMD/xAQEP8AAAD/AAAA/xISEv8QEBD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8IIiv/GCUp/xoaGv8WMDn/HTdA/zAwMP8tLS3/NDQ0/zMzM/8oKCj/Hx8f/yIiIv8nJyf/
                AAAA/wQEBP8eHh7/Kysr/xYWFv8VFRX/HBwc/y0tLf8VFRX/IiIi/yIiIv8SEhL/EhIS/wYGBv8rKyv/
                Hx8f/w8PD/8YGBj/IiIi/yQkJP8lJSX/EBAQ/yQkJP8GBgb/CgoK/wYGBv8HBwf/AwMD/wcHB/8AAAD/
                AAAA/w0NDf8GBgb/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/GCUp/xoaGv8WMDn/HTdA/zAwMP8zMzP/
                Ojo6/y4uLv8iIiL/Gxsb/ygoKP8fHx//AAAA/wYGBv8iIiL/ISEh/yUlJf8lJSX/Kioq/x4eHv8bGxv/
                NDQ0/ycnJ/8YGBj/GRkZ/wkJCf8rKyv/HBwc/xUVFf8SEhL/MTEx/xkZGf8rKyv/KCgo/yEhIf8YGBj/
                FRUV/x4eHv8iIiL/AwMD/x8fH/8AAAD/AAAA/yEhIf8eHh7/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/
                GCUp/xoaGv8WMDn/GjQ9/y0tLf8wMDD/Nzc3/zAwMP8oKCj/GRkZ/ygoKP8nJyf/AAAA/wYGBv8iIiL/
                NDQ0/wcHB/8SEhL/JScn/zMzM/8PEBD/KC0t/yUqKv8ZGRn/BgYG/wABAf8iJCT/ISEh/zc3N/8ZGRn/
                EhIS/xYWFv8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/GCUp/xoaGv8WMDn/HjhC/zQ0NP85OTn/PT09/zY2Nv8tLS3/
                Hx8f/zY2Nv80NDT/AAAA/wYGBv8AISH/ADAw/wAwMP8AKir/AB8f/wAWFv8AKCj/ABYW/wATE/8AIiL/
                ACsr/wAuLv8AExP/ABgY/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/GCUp/xoaGv8WMDn/
                HjhC/zY2Nv8zMzP/Li4u/0BAQP8xMTH/GRkZ/y0tLf8uLi7/AAAA/wYGBv8bGxv/JCQk/yQkJP8kJCT/
                JCQk/yQkJP8kJCT/Gxsb/ycnJ/8bGxv/Jycn/xsbG/8nJyf/Gxsb/yoqKv8bGxv/Kioq/xsbG/8qKir/
                Gxsb/ycnJ/8bGxv/Jycn/xsbG/8nJyf/Gxsb/yQkJP8kJCT/JCQk/yQkJP8kJCT/JCQk/xsbG/8nJyf/
                Gxsb/ycnJ/8bGxv/Jycn/xsbG/8qKir/Gxsb/xUVFf8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8IIiv/GCUp/xoaGv8WMDn/FzE6/y4uLv8oKCj/JSUl/zo6Ov8qKir/EBAQ/yQkJP8iIiL/
                AAAA/wQEBP8VFRX/GRkZ/yIiIv80NDT/JCQk/zQ0NP8YGBj/Jycn/ysrK/8nJyf/DAwM/wwMDP8qKir/
                KCgo/xMTE/8eHh7/Jycn/yIiIv8YGBj/JCQk/zExMf8wMDD/MTEx/zAwMP8xMTH/LS0t/zExMf8oKCj/
                LS0t/xwcHP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/CSQu/woXG/8MMD7/HTdA/zMzM/8uLi7/
                Kioq/zQ0NP8tLS3/GBgY/yoqKv8rKyv/AAAA/wYGBv8bGxv/LS0t/yoqKv8tLS3/Li4u/zMzM/8rKyv/
                ISEh/zc3N/8kJCT/MTEx/yUlJf82Njb/ISEh/zMzM/8oKCj/OTk5/ycnJ/8xMTH/JCQk/zk5Of8nJyf/
                OTk5/ycnJ/85OTn/KCgo/zQ0NP8wMDD/NDQ0/y0tLf8kJCT/JCQk/xsbG/8nJyf/Gxsb/ycnJ/8bGxv/
                Jycn/xsbG/8qKir/Gxsb/xUVFf8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/
                ABsl/wANEf8GKjj/HTdA/y4uLv8uLi7/Kioq/zk5Of8uLi7/GBgY/y4uLv8tLS3/AAAA/wYGBv8WFhb/
                FhYW/wYGBv8NDQ3/GRkZ/yIiIv8KCgr/Gxsb/x8fH/8YGBj/AwMD/wAAAP8fHx//DAwM/xgYGP8GBgb/
                GBgY/xAQEP8HBwf/Gxsb/yIiIv8EBAT/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/ACUz/wAYIf8GLDv/FS85/x4eHv8cHBz/HBwc/ycnJ/8ZGRn/
                EBAQ/x4eHv8fHx//AAAA/wQEBP8YGBj/JCQk/wYGBv8NDQ3/FhYW/x8fH/8KCgr/Gxsb/xgYGP8QEBD/
                BAQE/wAAAP8bGxv/EhIS/xsbG/8EBAT/HBwc/xkZGf8VFRX/GRkZ/yIiIv8EBAT/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IISv/ACg4/wJUdP8GJDD/
                DDJA/wciLP8HIiv/CCEr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/
                CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/
                CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/
                CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/
                CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/
                CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IISv/
                ByIr/wciLP8NLTn/CB4m/wgVGf8MKjb/Bik1/wAoOP8AJTP/ACY0/wANEf8ADRH/AA0R/wANEf8ADRH/
                AA0R/wANEf8ADRH/AA0R/wANEf8EHyn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/
                GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/
                GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/
                GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/
                GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/
                GCUp/xglKf8YJSn/GCUp/xglKf8XJCn/ACUz/wAoOP8IHib/Ghoa/xoaGv8WJiz/Bh8p/wJUdP8AGCH/
                ABoj/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8EERX/Ghoa/xoaGv8aGhr/
                Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/
                Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/
                Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/
                Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/
                Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/ABgh/wJUdP8IFRn/
                Ghoa/xgfIf8TNEH/CjND/wYkMP8GLDv/BjFB/wYgKf8GICn/BiAp/wYgKf8GICn/BiAp/wYgKf8GICn/
                BiAp/wYgKf8ILTr/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/
                FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/
                FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/
                FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/
                FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/
                FjA5/xYwOf8WLzn/Biw7/wYkMP8MKjb/FiYs/xM0Qf8SQVP/dBUY/09RYP95Cgz/gAIC/4ECAv+BAgL/
                ggMD/4ECAv+BAgL/gQIC/4ECAv+BAgL/ggMD/4ECAv+BAgL/gQIC/4ECAv+BAgL/ggMD/4ECAv+BAgL/
                gQIC/4ECAv+BAgL/ggMD/4ECAv+BAgL/gQIC/4ECAv+BAgL/ggMD/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                ggMD/4ECAv+BAgL/gQIC/4MGBv+BAgL/ggMD/4ECAv+BAgL/gQIC/4ECAv+CBAT/gwUF/4IEBP+BAgL/
                gQIC/4QICP+BAgL/ggMD/4IFBf+CBAT/gQIC/4EDA/+BAwP/ggMD/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                ggMD/4ECAv+BAgL/gQIC/4ECAv+BAgL/ggMD/4ECAv+BAgL/gQIC/4ECAv+BAgL/ggMD/4ECAv+BAgL/
                gQIC/4ECAv+BAgL/ggMD/4ECAv+BAgL/gQIC/4ECAv+BAgL/ggMD/4ECAv+AAgL/gAIC/4ACAv9/BAT/
                RlRk/wSryf9NKDD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/
                fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/
                fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/li8v/6RLS/+tXV3/smdn/6lUVP+sWVn/
                rFlZ/6xaWv+ePT3/oUVF/69hYf+wYmL/jh4e/6lUVP+rWFj/rVxc/7hzc/+0a2v/rFpa/61bW/+mTk7/
                fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/
                fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/fwAA/38AAP9/AAD/
                fwAA/38AAP9/AAD/fwAA/38AAP+BAgL/cQ4Q/0YnLv92Bwf/gAIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                gQIC/4MHB/+ECAj/hAgI/4MGBv+DBwf/gwcH/4MHB/+CBAT/hAkJ/4MHB/+ECAj/ggMD/4IEBP+DBgb/
                gwYG/4QICP+DBwf/gwcH/4MHB/+DBgb/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/
                gQIC/4ECAv+BAgL/gQIC/4ECAv+BAgL/gQIC/4ECAv+AAgL/gAIC/4ACAv9/BAT/Fj1M/yA7Rf8gOkT/
                HDU//x03Qf8fOUL/GTM8/x03Qf8fOUL/CiQt/wslL/8KJD7/CiQ//wokO/8KJDX/CiQv/wokM/8KJD//
                CiQ5/wokOP8KJD7/CiQ7/wokOf8KJDz/CiQz/wokPP8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/
                CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/
                CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/
                CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/
                CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KJC3/CiQt/wokLf8KIy3/CSQt/wkkLv8OMT7/
                CCo2/wciLP8LM0H/HTdA/zAwMP8tLS3/Kioq/zAwMP8eHh7/GBgY/yoqKv8rKyv/AAAA/wYGBv8KCiv/
                AAAP/wAAHv8AAAP/CQkP/wwMHP8AACL/Cgor/w0NGP8PDx//EBAn/w8PN/8JCSL/CQkY/wwMJP8PDw//
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8IISv/ACMx/wFObP8GIy7/GjQ9/y0tLf8tLS3/KCgo/zExMf8WFhb/
                EBAQ/yUlJf8wMDD/AAAA/wYGBv8lJSX/GRkZ/yUlJf8HBwf/NDQ0/yQkJP8GBgb/Gxsb/ygqKP8iIiL/
                Hh4e/xsbG/8bGxv/KCgo/zY2Nv8qKir/BgYG/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/ACAs/wATG/8GJzP/
                HjhC/zQ0NP8zMzP/Li4u/zY2Nv8wMDD/JSUl/yoqKv8xMTH/AAAA/wYGBv8tLS3/Jycn/yIiIv8kJCT/
                IiIi/zAwMP8GBgb/AAAA/wAoAP8AGQD/ABYA/wAkAP8ANAD/ACQA/wAzAP8DCQP/Hx8f/xUVFf8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8IIiv/FC44/xYjJ/8TNkT/HjhC/zQ0NP8zMzP/Li4u/zY2Nv8wMDD/JSUl/yoqKv8xMTH/
                AAAA/wYGBv8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wA8AP8AMAD/ABwA/wAAAP8AAAD/
                AC4A/wBCAP8ACQD/AAAA/wAAAP8AFQD/ACcA/wAtAP8AAAD/AAAe/wAALf8AADb/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAJ/8AADD/AAAl/wAALv8AABv/
                AAAS/wAAEv8AADP/AAAZ/wAALf8AABj/AAAA/wAAAP8AAB7/AAAr/wAAIv8AACX/AAAu/wAAAP8AFgD/
                AEAA/wANAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAZAP8IOiv/GCUp/xoaGv8WMDn/HjhC/zQ0NP8zMzP/
                Li4u/zY2Nv8wMDD/JSUl/yoqKv8xMTH/AAAA/wYGBv8xMTH/MTEx/ygoKP8wMDD/NjY2/zQ0NP8TExP/
                AwMD/zo6Ov8wMDD/FhYW/x4eHv8AAAD/Kioq/yUlJf8qKir/Pz8//w0NDf8TExP/PUU9/xMTE/8AAAD/
                FhYW/yUlJ/8kJCr/Kioq/yIiIv8TExP/KCgo/y4uLv8PDw//JSUl/woKCv88PDz/JCQk/y0tLf8cHBz/
                AAAA/xYWFv8AAAD/AAAA/wAAAP8AAAD/AAAA/xISEv8zMzP/GRkZ/y0tLf8YGBj/FhYW/wAAAP8qKir/
                JSUq/ygoKP8AAAD/AAAA/wAAAP8cHBz/JSUl/x8fH/8nJyf/Hx8f/xYWFv8HBwf/GRkZ/x4eHv8qKir/
                IiIi/y4uLv8HBwf/GBgY/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/
                GCUp/xoaGv8WMDn/HTdA/zAwMP8uLi7/Kioq/zAwMP8qKir/Hx8f/yUlJf8rKyv/AAAA/wQEBP8tLS3/
                Kysr/yQkJP8rKyv/MTEx/zAwMP8KCgr/DQ0N/zY2Nv8rKyv/FRUV/xsbG/8AAAD/JSUl/yEhIf8nJyf/
                OTk5/woKCv8SEhL/Nzc3/xISEv8AAAD/ExMT/x8fH/8hISH/JSUl/x8fH/8SEhL/JSUl/zAwMP8NDQ3/
                ISEh/wkJCf83Nzf/Hx8f/yoqKv8ZGRn/AAAA/xUVFf8AAAD/AAAA/wAAAP8AAAD/AAAA/xAQEP8uLi7/
                GBgY/ygoKP8VFRX/FRUV/wAAAP8lJSX/ISEh/yQkJP8AAAD/AAAA/wAAAP8YGBj/LS0t/x4eHv8nJyf/
                Gxsb/y4uLv8ZGRn/Gxsb/yIiIv8rKyv/HBwc/xUVFf8bGxv/Dw8P/xISEv8ZGRn/HBwc/yIiIv8fHx//
                GRkZ/xsbG/8fHx//AAAA/wAAAP8IIiv/GCUp/xoaGv8WMDn/HTdA/zAwMP80NDT/MTEx/zAwMP8lJSX/
                JCQk/y0tLf8tLS3/AAAA/wYGBv8oKCj/IiIi/x4eHv8iIiL/ISEh/x8fH/8KCgr/BAQE/xA0EP8NFg3/
                CR4J/wwYDP8NDQ3/FhYW/w8PD/8NDQ3/FRUV/wYGBv8HBwf/EBAQ/wcHB/8AAAD/CQkJ/xISEv8JCQn/
                DQ0N/wwMDP8EBAT/DAwM/xsbG/8EBAT/Dw8P/wQEBP8QEBD/Dw8P/w8PD/8HBwf/AAAA/woKCv8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wYGBv8VFRX/BwcH/xISEv8JCQn/CgoK/wAAAP8NDQ3/Dw8P/w0NDf8AAAD/
                AAAA/wAAAP8KCgr/Dw8P/w8PD/8QEBD/BgYG/xUVFf8JCQn/Dw8P/xYWFv8hISH/DQ0N/xAQEP8KCgr/
                BwcH/wkJCf8PDw//CgoK/xMTE/8NDQ3/EBAQ/xYWFv8WFhb/AAAA/wAAAP8IIiv/GCUp/xoaGv8WMDn/
                HTdA/zAwMP8xMTH/NjY2/zMzM/8lJSX/JCQk/zMzM/8xMTH/AAAA/wQEBP8qKir/ISEh/xwcHP8hISH/
                Hh4e/ysrK/8EBAT/AAAA/wAlAP8AIgD/ADEA/wAfAP8NGA3/BxAH/wAKAP8DAwP/DQ0N/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8IIiv/GCUp/xoaGv8WMDn/HTdA/zAwMP82Njb/Nzc3/zk5Of8QEBD/IiIi/ygoKP8kJCT/
                AAAA/wYGBv8rKyv/Jycn/yIiIv8iIiL/IiIi/zAwMP8GBgb/AAAA/wArAP8AJAD/ABkA/wATAP8AJwD/
                ADEA/wAtAP8GDwb/ISEh/wEBAf8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/GCUp/xoaGv8WMDn/GjQ9/y0tLf82Njb/
                NDQ0/zExMf8lJSX/GBgY/xYWFv8nJyf/AAAA/wYGBv8kJCT/Jycn/yIiIv8iIiL/IiIi/zAwMP8GBgb/
                AAAA/wAlAP8ALQD/ACgA/wADAP8AJwD/ACEA/wA5AP8AGQD/GBgY/wwMDP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/
                GCUp/xoaGv8WMDn/HjhC/zQ0NP86Ojr/Pz8//zk5Of8kJCT/JSUl/ycnJ/8rKyv/AAAA/wYGBv8tLS3/
                Jycn/yIiIv8kJCT/IiIi/zAwMP8GBgb/AAAA/wA0AP8AJwD/ABYA/wAeAP8AIQD/ADEA/wAtAP8DDAP/
                MTEx/w0NDf8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/GCUp/xoaGv8WMDn/HjhC/zQ0NP85OTn/PT09/zc3N/8wMDD/
                Jycn/ygoKP8rKyv/AAAA/wYGBv8tLS3/Jycn/yIiIv8kJCT/IiIi/y0tLf8GBgb/AAAA/wA0AP8AJwD/
                ABYA/wAeAP8AIQD/AC4A/wAtAP8DDAP/Kysr/woKCv8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/GCUp/xoaGv8WMDn/
                FzE6/ysrK/8wMDD/NDQ0/ysrK/8qKir/GBgY/ysrK/8qKir/AAAA/wQEBP8iIiL/ISEh/xkZGf8bGxv/
                GRkZ/ygoKP8EBAT/AAAA/wAtAP8AEgD/ACQA/wAVAP8PDw//EhkS/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8IIiv/CSQu/woXG/8MMD7/HTdA/zAwMP82Njb/Nzc3/zExMf8iIiL/Hh4e/zAwMP8xMTH/
                AAAA/wYGBv8zMzP/ISEh/yIiIv8TExP/Dw8P/xUqFf8DJAP/ABMA/wAkAP8AEwD/AEAA/wAVAP8HIQf/
                DBsM/wwMDP8PDw//AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/ABsl/wANEf8GKjj/HTdA/y4uLv8rKyv/
                MDAw/zMzM/8tLS3/GBgY/y4uLv8uLi7/AAAA/wQEBP8lJSX/JCQk/xkZGf8WFhb/EBAQ/w0cDf8BFgH/
                ABMA/wAuAP8AEAD/AB4A/wAVAP8ANgD/ABkA/wwPDP8iIiL/Gxsb/wYGBv8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8IIiv/
                ACs7/wAfK/8GMkL/FS85/x4eHv8cHBz/HBwc/ycnJ/8ZGRn/EBAQ/x4eHv8eHh7/AAAA/wQEBP8cHBz/
                Gxsb/xYWFv8VFRX/FRUV/yIiIv8EBAT/AAAA/wAbAP8AEwD/ABYA/wAHAP8AGwD/ABkA/wAVAP8ZGRn/
                IiIi/wcHB/8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8IISv/ACg4/wJUdP8GJDD/DDJA/wciLP8HIiv/CCEr/wgiK/8IIiv/
                CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/
                CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/
                CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/
                CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/
                CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/
                CCIr/wgiK/8IIiv/CCIr/wgiK/8IIiv/CCIr/wgiK/8IISv/ByIr/wciLP8NLTn/CB4m/wgVGf8MKjb/
                Bik1/wAoOP8AJTP/ACY0/wANEf8ADRH/AA0R/wANEf8ADRH/AA0R/wANEf8ADRH/AA0R/wANEf8EHyn/
                GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/
                GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/
                GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/
                GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/
                GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8YJSn/GCUp/xglKf8XJCn/
                ACUz/wAoOP8IHib/Ghoa/xoaGv8WJiz/Bh8p/wJUdP8AGCH/ABoj/wAAAP8AAAD/AAAA/wAAAP8AAAD/
                AAAA/wAAAP8AAAD/AAAA/wAAAP8EERX/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/
                Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/
                Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/
                Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/
                Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/
                Ghoa/xoaGv8aGhr/Ghoa/xoaGv8aGhr/ABgh/wJUdP8IFRn/Ghoa/xgfIf8TNEH/CjND/wYkMP8GLDv/
                BjFB/wYgKf8GICn/BiAp/wYgKf8GICn/BiAp/wYgKf8GICn/BiAp/wYgKf8ILTr/FjA5/xYwOf8WMDn/
                FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/
                FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/
                FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/
                FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/
                FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WMDn/FjA5/xYwOf8WLzn/Biw7/wYkMP8MKjb/
                FiYs/xM0Qf8SQVP/
            </Image>
        </BitMap>
        <Widget index="0" type="434f4e53" display_nr="1" y="4" anz_rows="4" anz_cols="16">
            <Data>
                <Data>13</Data>
                <Data>3</Data>
                <Data>0</Data>
                <Data>0</Data>
            </Data>
        </Widget>
        <Widget index="1" type="44425547" display_nr="1" has_focus="true" has_scrollfocus="true" anz_rows="4" anz_cols="16">
            <Data>
                <Data>0</Data>
                <Data>0</Data>
                <Data>0</Data>
                <Data>0</Data>
            </Data>
        </Widget>
    </View>
</MA>]]
        
        -- File configuration
        local dir = {}
        dir.main = gma.show.getvar('PATH')..'/'
        dir.importexport = dir.main..'importexport/'
        
        fn = {}
        fn.temp = [[_templ_view.xml]]
        
        file = {}
        file.temp = dir.importexport..fn.temp
        
        -- Write temp file
        local writefile = io.open(file.temp, 'w')
        writefile:write(cmd_script)
        writefile:close()
    
        -- Import and call view
        local view_slot = _M.poolitem.advanceSpace('View', 1, 1)
        gma.cmd('SelectDrive 1')
        gma.cmd('Import "'..fn.temp..'" View '..view_slot)
        gma.cmd('View '..view_slot..' /screen='..tostring(screen)) 
        
        -- Cleanup
        os.remove(file.temp)
        gma.cmd('Delete View '..view_slot)
    end
} --end of gui table





---- Console Status

_M.dmx = {        --sub-module v1.0
    --split a string to two table values
    split_addr = function(str, numindex)
      if type(str) == 'number' then str = tostring(str) end
      local index = str:find(':')
      if not index then
        index = str:find('[.]')
      end
      local t = {}

      if index then
        if numindex then
            t[1] = tonumber(str:sub(1, index-1))
            t[2] = tonumber(str:sub(index+1, #str))
            return t
        else
            t.univ    = tonumber(str:sub(1, index-1))
            t.ch    = tonumber(str:sub(index+1, #str))
        end
      else
          error('invalid address provided; no separator found')
      end
      
      return t
    end,


    merge_addr = function(list, seperator)
        seperator = seperator or '.'
        
        univ = list.univ or list[1]
        ch = list.ch or list[2]
        univ, ch = tostring(univ), tostring(ch)
        
        --make ch 3 digits long
        while #ch < 3 do
            ch = '0'..ch
        end
        
        --join strings
        local t = univ..seperator..ch
        return t
    end,

    univ = {},

    univ_add = function(self, univ_num)
        if not _M.csv.objToArray then error('Required function not found: csv.objToArray(str:obj)') end
        if not _M.gui.superBar then error ('Required module not found: gui.superBar') end
        
        --import new universe to CSV with progress bar displayed
        pBar_add = _M.gui.superBar:new('Importing Universe '..univ_num)
        pBar_add:setrange(0, 512)
        self.univ[univ_num] = _M.csv.objToArray('DmxUniverse '..univ_num..'.*', pBar_add)    --import new universe
        pBar_add:stop()
        
        --[[
        for i = 1, 512 do
            self.univ[univ_num][i].addr = self.split_addr(self.univ[univ_num][i].addr, true)
        end
        --]]
    end,
    
    check_space = function(self, addr_start, spaces)
        -- if address provided in string format, convert to table object
        if spaces > 512 then spaces = 512 end    --idiot-protection
        if type(addr_start) == 'string' then addr_start = self.split_addr(addr_start) end    
        local addr_current = {univ = addr_start.univ or addr_start[1], ch = addr_start.ch or addr_start[2]}
        
        
        local confirmed = false
        local t = {ct = 0}
        
        while not confirmed do
            -- if the universe to be checked is not currently in memory, import it
            if not self.univ[addr_current.univ] then
                self:univ_add(addr_current.univ)
                --self.univ[addr_current.univ] = self:univ_add(addr_current.univ)
            end
            

            -- scan for empty channels
            for i = addr_current.ch, 512 do
                if #tostring(self.univ[addr_current.univ][i].fixture) == 0 then                    --if the line has no fixture assigned
                    t.ct = t.ct + 1
                    if not t.addr then
                        t.addr = {univ = addr_current.univ, ch = addr_current.ch}
                    end
                    if t.ct == spaces then
                        confirmed = true
                        break
                    end
                    addr_current.ch = i + 1
                else                                                                            --if the line DOES have a fixture assigned
                    t.ct = 0
                    t.addr = nil
                    addr_current.ch = i + 1
                end
            end
            
            -- if not found in that universe, reset parameters and start again at next universe number
            if not confirmed then 
                addr_current = {univ = addr_current.univ + 1, ch = 1}
                t.ct = 0
                t.addr = nil
            end        --if end of universe is reached with no match for spaces
        end

        return t.addr
    end


}
--[[ ----TEST----
---- TEST ----
return function()
    gma.echo('final result: '..dmx.merge_addr(dmx:check_space('50.1', 2)))
end
--]]


_M.env = {        -- environment module v1.0.0
    getprog = function() 
        --returns 'live', 'blind', or 'preview'

      local file = {}
      file.name = 'tempfile_programmercheck.csv'
      file.directory = gma.show.getvar('PATH')..'/reports/'
      file.fullpath = file.directory..file.name
      
      gma.cmd('SelectDrive 1')                             --select internal drive
      gma.cmd('List Programmer /f=\"'..file.name..'\"') --create temp file
      
      local prog
      for line in io.lines(file.fullpath) do        --check line for status as live, blind, or preview programmer
        if     line:find('Programmer') == 1   then prog = 'live'
        elseif line:find('BlindProgrammer')   then prog = 'blind'
        elseif line:find('PreviewProgrammer') then prog = 'preview' end
      end
      os.remove(file.fullpath)                         --delete temp file
      
      return prog
    end,
    
    self_id = function(self_table)    -- function v1.2.0
        -- takes self table (table captured by {...}) and returns pool item number where executing plugin is currently stored
        
        -- self_table has to be extracted before "return" portion of plugin.
        -- (AKA it must be generated on plugin load)
        
        local unpack = table.unpack

        local self = self_table[2]
        local indices = {0, 0}

        repeat
            indices = {self:find('%d+', indices[2] + 1)}
        until indices[2] == #self

        local self_id = tonumber(self:sub(unpack(indices)))
        return self_id
    end,

    rootnum = function(str) -- function v1.0.0
    -- retrieves directory number for a given object-type string (i.e. '8' for images)
    -- CAREFUL WHAT STRING YOU SUPPLY
        
        local unpack = table.unpack
        str = str:lower()                                                                --set string to find to lowercase to avoid case errors
        
        -- tempfile info
        local temp = {}
        temp.fn = 'tempfile_getroot.csv'
        temp.dir = gma.show.getvar('PATH')..'/reports/'
        temp.file = temp.dir..temp.fn
        
        -- retrieve info from showfile
        gma.cmd('SelectDrive 1')
        gma.cmd('List /f=\"'..temp.fn..'\"')
        
        local line_locate = 0
        for line in io.lines(temp.file) do
            if line:lower():find(str) then                                                --if there is a lowercase match 
                line_locate = tonumber(line:sub(unpack({line:find('%d+')})))            --retrieve the number following it, converted to number format
                break
            end
        end
        
        os.remove(temp.file)                                                            --remove temp file from system
        
        -- return result
        if line_locate == 0 then
            return false, 'string not found'
        else
            return line_locate
        end    
    end,
    
    }





---- File/Data Manipulation ----


_M.data = {
    sf = {-- SaveFile module v0.0.1
        header = '--[[',
        footer = '--]]',
        
        save = function(config)
            -- config.table, .name, .num
            
            local l_dir = _M.data.sf    -- set "local directory" (for if module is moved)
            
            local script_tbl = {l_dir.header, _M.table.tostring(config.table), l_dir.footer}
            local script = table.concat(script_tbl, '\n')
            
            
            gma.cmd('Unlock Plugin '..config.num)
            _M.plugin.write({name = config.name, num = config.num, script = script}, true)
            gma.cmd('Lock Plugin '..config.num)
        end,
        
        
        read = function(plugin_num)
            -- function to prep header/footer tags for search
            
            local l_dir = _M.data.sf
            
            local function s_prep(str)    -- string search prep function
                return (str:gsub('(%W)', '%%%1'))
            end
            
            local script = _M.plugin.read(plugin_num)                                                    -- extract script from plugin
            local script_tbl = script:gsub(s_prep(l_dir.header)..'(.+)'..s_prep(l_dir.footer), '%1')    -- extract 
            
            local tbl = _M.table.fromstring(script_tbl)
            return tbl
        end
    },
    
    parseColon = function(info_raw)
    -- takes raw string in [key: value] format with one entry pair per line and returns as table

        local t = {}

        for i in info_raw:gmatch('.-%:.-[\n]') do
            local k, v = i:match('%s*([^\n]+)%s*%:%s*([^\n]+)[%s\n]*')
            if k and v then
                t[k] = v
            end
        end

        -- local finish = info_raw:match('[\n][^\n]-%:[^\n]-$')
            -- local k, v = finish:match('%s*(.+)%s*%:%s*(.+)[%s\n]*')
            -- t[k] = v

        return t
    end,

    trunc = function(num, modulus)
        return (num - num%modulus)
    end,    
    
    sr_ct = 1,    -- _M.data.sr_ct
    
    setrange = function(config)
        -- config input:
        --\\ .size .start .loops .direct
        -- output format
        --\\ .size (.start): {.start .finish}
        --\\ \\ +.direct: start, finish
        --\\ .size .loops:   start1, start2, start3, ...
        
        
        local l = _M.data
        
        if config.start then l.sr_ct = config.start end

        if (not config.loops) then
            local start = l.sr_ct
            local finish = l.sr_ct + config.size - 1
            l.sr_ct = l.sr_ct + config.size
            
            if config.direct then return start, finish
            else return {start = start, finish = finish} end
        
        elseif config.loops then
            local results = {}
            for i = 1, config.loops do
                results[i] = l.sr_ct
                l.sr_ct = l.sr_ct + range
            end
            return unpack(results)
        end
        
    end
}



_M.table = {    -- table storage/loading Module v1.0.0
    tostring = function(tbl)
      -- function found online at http://lua-users.org/wiki/SaveTableToFile
      local function exportstring( s )
        return string.format("%q", s)
      end

      local charS,charE = "   ","\n"                -- space char, newline char
      local build_t,err = _M.class.superTable:new()        -- define "build_t" to append to

      -- initiate variables for save procedure
      local tables,lookup = { tbl },{ [tbl] = 1 }    -- DO NOT UNDERSTAND LOOKUP VARIABLE
      build_t:append( "return {"..charE )

      for idx,t in ipairs(tables) do
         build_t:append( "-- Table: {"..idx.."}"..charE )
         build_t:append( "{"..charE )
         local thandled = {}

         for i,v in ipairs( t ) do
            thandled[i] = true
            local stype = type( v )
            -- only handle value
            if stype == "table" then
               if not lookup[v] then
                  table.insert( tables, v )
                  lookup[v] = #tables
               end
               build_t:append( charS.."{"..lookup[v].."},"..charE )
            elseif stype == "string" then
               build_t:append(  charS..exportstring( v )..","..charE )
            elseif stype == "number" then
               build_t:append(  charS..tostring( v )..","..charE )
            end
         end

         for i,v in pairs( t ) do
            -- escape handled values
            if (not thandled[i]) then

               local str = ""
               local stype = type( i )
               -- handle index
               if stype == "table" then
                  if not lookup[i] then
                     table.insert( tables,i )
                     lookup[i] = #tables
                  end
                  str = charS.."[{"..lookup[i].."}]="
               elseif stype == "string" then
                  str = charS.."["..exportstring( i ).."]="
               elseif stype == "number" then
                  str = charS.."["..tostring( i ).."]="
               end

               if str ~= "" then
                  stype = type( v )
                  -- handle value
                  if stype == "table" then
                     if not lookup[v] then
                        table.insert( tables,v )
                        lookup[v] = #tables
                     end
                     build_t:append( str.."{"..lookup[v].."},"..charE )
                  elseif stype == "string" then
                     build_t:append( str..exportstring( v )..","..charE )
                  elseif stype == "number" then
                     build_t:append( str..tostring( v )..","..charE )
                  end
               end
            end
         end
         build_t:append( "},"..charE )
      end
      build_t:append("}")
      
      return build_t:concat()
    end,

    fromstring = function(tbl_string)
      -- -- function found online at http://lua-users.org/wiki/SaveTableToFile
      local ftables,err = load(tbl_string)    -- load string
      if err then return _,err end            -- return error if present
      local tables = ftables()                -- execute string
      for idx = 1,#tables do
         local tolinki = {}
         for i,v in pairs( tables[idx] ) do
            if type( v ) == "table" then
               tables[idx][i] = tables[v[1]]
            end
            if type( i ) == "table" and tables[i[1]] then
               table.insert( tolinki,{ i,tables[i[1]] } )
            end
         end
         -- link indices
         for _,v in ipairs( tolinki ) do
            tables[idx][v[2]],tables[idx][v[1]] =  tables[idx][v[1]],nil
         end
      end
      return tables[1]
    end,

    buildpath = function(master, list)
        local dir_c = master
        for _, v in ipairs(list) do
            if not dir_c[v] then
                dir_c[v] = {}
                dir_c = dir_c[v]
            end
        end
        
        return dir_c
    end,
    
    keyswap = function(tbl)        -- takes table of keys set to true and returns list with those keys as values
        local t = ST:new()
        for k, _ in pairs(tbl) do
            t:append(k)
        end
        
        return t
    end,
    
    copy = function(obj)
        if type(obj) ~= 'table' then return obj end
        local t = {}
        for k, v in pairs(obj) do t[k] = _M.table.copy(v) end
        return t
    end
}



_M.xml = {    --sub-module v1.0
    sub = function(script, sublist)
        --sub format = list.(substitution variable) = (substitution value)
        -- replaces <!(substitution variable> with value in script
        
        for k, v in pairs(sublist) do
            script = script:gsub('<!'..k..'>', v)
        end
        
        return script
    end,
    

    extract_vartags = function(script)
        local t = {}
        local indices = {0, 0}
        local remaining = true
        while remaining do
            indices = {script:find('<![^-].->', (indices[2] + 1))}
            if not indices[2] then break
            elseif indices[2] == #script then remaining = false end
            t[#t+1] = script:sub(indices[1] + 2, indices[2] - 1)
        end

        return t
    end
}



_M.csv = {    --sub-module version 1.1.0
    objToArray = function(obj, pBar)

    -------- NESTED FUNCTIONS --------

        ---- converts a CSV line to a number-indexed array ----
        local function CSVlineToArray(str)
          local t = {}
          local indices = {1, 1}
          local remaining = true
          local restart = 1
          while remaining == true do
            local pt = {str:find(',', restart)}        --find next comma at or after restart point
            if pt[1] then
              t[#t+1] = str:sub(restart, pt[1]-1)    --extract string from restart point to one position before comma
              restart = pt[2] + 1                     --reset restart point for one position after most recent found comma
            else
              remaining = false
              t[#t+1] = str:sub(restart, str[-1])
            end
          end

          return t
        end


        ---- Converts a CSV line to an array using the indices of an input header table ----
        local function CSVsubheaderToArray(str, header)
          local t = {}
          local indices = {1, 1}
          local remaining = true
          local restart = 1
          local index = 1
          while remaining == true do
            local pt = {str:find(',', restart)}        --find next comma at or after restart point
            if pt[1] then
              t[header[index]] = str:sub(restart, pt[1]-1)    --extract string from restart point to one position before comma
              restart = pt[2] + 1                     --reset restart point for one position after most recent found comma
              index = index + 1
            else
              remaining = false
              t[header[index]] = str:sub(restart, str[-1])
            end
          end

          return t
        end

        ---- takes a CSV file, uses the first row as the headers , and indexes the contents into a table using those headers 
        local function CSVfileToArray(file, pBar)
            local header = {}
            local results = {}

            local firstLine = true
            for line in io.lines(file) do
              if firstLine then
                header = CSVlineToArray(line)
                for k, v in pairs(header) do header[k] = v:lower() end
                firstLine = false
              else
                results[#results+1] = CSVsubheaderToArray(line, header)
                if pBar then pBar:set(1, true) end
              end
            end

            return results
          
        end
        
        
        
        ---- main function ----
        -- filepaths
        local dir    = {temp = gma.show.getvar('PATH')..'/reports/'}
        local fn     = {temp = 'export_csvobject.csv'}
        local file    = {temp = dir.temp..fn.temp}
        
        -- export object
        gma.cmd('SelectDrive 1')
        if obj then    gma.cmd('List '..obj..' /f=\"'..fn.temp..'\"')
        else gma.cmd('List /f=\"'..fn.temp..'\"') end
        
        local t        --declare before if statement
        -- check if file exists; if not, no object was found for export
        local test = io.open(file.temp)
        if test then                                    --if file is present
            test:close()                                --close test file
            
            -- convert into table
            t = CSVfileToArray(file.temp, pBar)
                
            -- delete temp file
            os.remove(file.temp)
        else
            t = {}
        end

        
        -- return result
        return t
    end,
    

    
    ---- function to take a two-dimensional table and convert it to a CSV text string
    write_table = function(list)
        local t = ''
        for i = 1, #list do
          t = t..table.concat(list[i], ',')..'\n'
        end

        return t
    end
}


_M.lua = {
    sub = function(script, sublist)
        -- end-tag
        -- <!end-replacements> --
        
        
        --sub format = list.(substitution variable) = (substitution value)
        -- replaces <!(substitution variable> with value in script
        -- auto-formats strings and places between double quotes

        -- for strings, use {yourstring, true} to use literal-brackets
        -- or {yourstring} or {yourstring, nil, true} to have a string implemented directly as is,
        -- with no quote formatting or brackets

        local function s_prep(str)    -- format string for search
            return (str:gsub('(%W)', '%%%1'))
        end

        -- nested function to determine appropriate number of brackets to use on quoted
        -- strings containing brackets themselves
        local function literal_bracket(str)
            local tbl = {}
            for match in str:gmatch('%[[=]*%[') do
                tbl[#match-2] = true
            end

            local x = 0
            while true do
                if not tbl[x] then break end
                x = x + 1
            end
            print(x)

            local rep = string.rep('=', x)

            return '['..rep..'[', ']'..rep..']'
        end

        local end_tag = '<!end-replacements>'
        end_tag = s_prep(end_tag)

        local head, body = script:match('(.-'..end_tag..')(.*)') --separate replacement zone from rest of script

        for k, v in pairs(sublist) do
            local value = v
            if type(v) == 'string' then value = string.format("%q", v)
            elseif type(v) == 'table' then
                if v[2] and (type(v[1]) == 'string') then
                    local b = {literal_bracket(v[1])}
                    value = b[1]..v[1]..b[2]
                else
                    value = v[1]
                end
            end

            head = head:gsub('<!'..k..'>', value)
        end

        return head..body
    end
}


    
---- Object Properties
    
_M.get = {
    handle = gma.show.getobj.handle
    ,
    class = function(obj)
        gma.echo('ignore potential "error: returns string:classname" message below')
        return gma.show.getobj.class(_M.get.handle(obj))
    end
    ,
    label = function(obj)
        return gma.show.getobj.label(_M.get.handle(obj))
    end
    ,
    name = function(obj)
        local t = gma.show.getobj.name(_M.get.handle(obj))
        if not t then t = gma.show.getobj.label(_M.get.handle(obj)) end
        return t
    end
    ,
    verify = function(obj)
        gma.echo('ignore potential "error: returns bool:result" message below')
        return gma.show.getobj.verify(_M.get.handle(obj))
    end
    ,
    amount = function(obj)
        return gma.show.getobj.amount(_M.get.handle(obj))
    end
    ,
    parent = function(obj)
        return gma.show.getobj.amount(_M.get.handle(obj))
    end
    ,
    number = function(obj)
        return gma.show.getobj.number(_M.get.handle(obj))
    end
    ,
    index = function(obj)
        return gma.show.getobj.index(_M.get.handle(obj))
    end
    ,
    child = function(obj, index)
        return gma.show.getobj.number(_M.get.handle(obj), index)
    end,

    ---- more involved than MA shortcuts ----
    
    appearance = function(obj, convertPercent)
    -- input - obj:str, (convertPercent:bool)
    
    -- output from aqua-colored object
    --\\ (convertPercent true)    return 0  100  50
    --\\ (convertPercent false)    return 0  255  127
    
    
    
        local l_dir = _M.get        -- local "directory"
    
        ---- Nested Functions ----
        function math.round(num)
          if num - num%1 < 0.5 then
            return math.floor(num)
          else
            return math.ceil(num)
          end
        end

        function match(a, b) --checks to see if a and b match or (if strings) if one contains the other
            if a == b then return true;
            elseif type(a) == 'string' and type(b) == 'string' then
                if a:lower():find(b:lower()) or b:lower():find(a:lower()) then return true
                else return false end
            else return false
            end
        end

        function table.find(t, target, i, j)  --searches table for target text using user-defined [match] function
            local i = i or 1
            local j = j or #t
          

            for n = i, j do
                if match(t[n], target) == true then    return n end
            end
            return nil
        end


        ---- Filesystem Info ----
        local folder_options = {'effects', 'macros', 'masks', 'matricks', 'user_profile', 'plugins'}
        local poolType
        local t = table.find(folder_options, obj:sub(1,4))  --check if argument has a predefined export folder in MA (search using first 4 letters of input)
        if t then poolType = folder_options[t]
        else poolType = 'importexport' end
        
      
      
        local fn = {}
            fn.temp = '_tempfile_getappearance'
        local dir = {}
            dir.main = gma.show.getvar('PATH')..'/'
            dir.obj  = dir.main..poolType..'/'
        local file = {}
            file.xml = dir.obj..fn.temp..'.xml'
            file.lua = dir.obj..fn.temp..'.lua'
        
      
        ---- Main Function ----
      
        --check if object actualy exists
        local verified = _M.get.verify(obj)
        if not verified then
            return nil, 'Error: object does not exist.' end
            
        -- export object
        gma.cmd('Export '..obj..' \"'..fn.temp..'\"')

        -- convert XML file into a table
        local t = {}  
        for line in io.lines(file.xml) do
            t[#t + 1] = line
        end
      
        -- remove temp file(s)
        os.remove(file.xml)
        if poolType == 'plugins' then  -- delete LUA file if export item was a plugin
            os.remove(file.lua)
        end

        local colorHex
        for i = 1, #t do
            if t[i]:find('Appearance Color') then
                local indices
                while true do
                    local j = j
                    indices = {t[i]:find('\"%x+\"')}      --find color hex code
                    if indices[2]-indices[1]-1 == 6 then  --if length matches hex color code length
                        indices[1], indices[2] = indices[1] + 1, indices[2] - 1 --reset index values to start and end of actual hex string
                        break
                    elseif indices then
                        j = indices[2]
                    else
                        gma.echo('error: appearance function triggered, not found')
                        return nil
                    end
                end
                colorHex = t[i]:sub(indices[1], indices[2])
                break
            end
        end
      
        local colorTable = {}  --convert string pairs into numbers in a table - red, green, then blue
        colorTable[1] = tonumber('0x'..colorHex:sub(1,2))
        colorTable[2] = tonumber('0x'..colorHex:sub(3,4))
        colorTable[3] = tonumber('0x'..colorHex:sub(5,6))
      
        if convertPercent then
            for i = 1, #colorTable do
                colorTable[i] = math.round(colorTable[i]/2.55)
            end
        end
        
        return table.unpack(colorTable)
    end,

    list = function(object)
        -- filepath info --
        local location = 'reports'
        
        local fn = {temp = 'temp_generatepoolobject.csv'}
        
        local dir = {}
            dir.main = gma.show.getvar('PATH')..'/'
            dir.temp = dir.main..location..'/'

        local file = {temp = dir.temp..fn.temp}

        
        -- get script --
        gma.cmd('List '..object..' /f="'..fn.temp..'"')            -- export object
        io.input(file.temp)                                            -- set input file to generated script
        local script = io.read('*a')                                -- load full script as string
        
        os.remove(file.temp)                                        -- delete temp file (does not appear to be working currently)
        
        
        -- return --
        return script
    end,
    
    execlist = function(timecode)                                            -- returns list of executors contained in timecode pool item(s)
        local scr = _M.get.script('Timecode '..timecode)
        local e_list = {}
        for match in scr:gmatch('<Object name=\".-\">') do
            local num = match:match('(%d+%.%d+)\">$')
            if num then    e_list[num] = true end
        end
        e_list = _M.table.keyswap(e_list)
        e_list.numstr = e_list:concat(' + ')
        e_list.str = 'Executor '..e_list.numstr
        
        
        return e_list
    end,
    
    seqlist = function(execs)
        -- input ex: 'Executor 15'; 'Page 1 thru 4 Executor 1 + 13 Thru 15'
        if execs:lower():find('^%s*page') == 0 and execs:lower():find('^%s*executor') == 0 then
            execs = 'Executor '..execs end
        local scr = _M.get.list(execs)
        local seq_list = {}
        for match in scr:gmatch('Sequence=Seq %d+') do
            seq_list[match:match('%d+')] = true
        end
        seq_list = _M.table.keyswap(seq_list)
        seq_list.numstr = seq_list:concat(' + ')
        seq_list.str = 'Sequence '..seq_list:concat(' + ')
        
        return seq_list
    end,
    
    script = function(object, alternate_location)
        -- filepath info --
        local location = alternate_location or 'importexport'
        
        local fn = {temp = 'temp_generatepoolobject.xml'}
        
        local dir = {}
            dir.main = gma.show.getvar('PATH')..'/'
            dir.temp = dir.main..location..'/'

        local file = {temp = dir.temp..fn.temp}

        
        -- get script --
        gma.cmd('SelectDrive 1')
        gma.cmd('Export '..object..' "'..fn.temp..'" /nc')            -- export object
        io.input(file.temp)                                            -- set input file to generated script
        local script = io.read('*a')                                -- load full script as string
        
        os.remove(file.temp)                                        -- delete temp file (does not appear to be working currently)
        
        
        -- return --
        return script
    end,
    
    grpfixt = function(grpNum)    --function v1.1.0
      gma.cmd('SelectDrive 1') --select the internal drive
      
      local file = {}
      file.name = 'tempfile.xml' 
      file.directory = gma.show.getvar('PATH')..'/'..'importexport'..'/'
      file.fullpath = file.directory..file.name
      
      gma.cmd('Export Group ' .. grpNum .. ' \"' .. file.name .. '\"') -- create temporary file
      
      local t = {}  --convert XML file into a table
      for line in io.lines(file.fullpath) do
        t[#t + 1] = line
      end
      os.remove(file.fullpath) --delete temporary file
      
      local groupList = {} --declare groupList
      for i = 1, #t do
        if t[i]:find('Subfixture ') then
          local indices = {t[i]:find('\"%d+\"')} --find points of quotation marks
          indices[1], indices[2] = indices[1] + 1, indices [2] - 1 --move reference points to first an last characters inside those marks
          
          --label based on status as a fixture or as a channel
          local fixture
          if t[i]:find('fix_id') then
            fixture = 'Fixture ' .. tostring(t[i]:sub(unpack(indices))) --extract the number as the fixture number
          elseif t[i]:find('cha_id') then
            fixture = 'Channel ' .. tostring(t[i]:sub(unpack(indices))) end --extract the number as the fixture number
            
          --if the object contains a subfixture...
          if t[i]:find('sub_index') then
            local indices = {t[i]:find('\"%d+\"', indices[2]+2)}
            indices[1], indices[2] = indices[1] + 1, indices[2] - 1
            fixture = fixture .. '.' .. tostring(t[i]:sub(unpack(indices)))
          end
          
          --append to list
          groupList[#groupList + 1] = fixture --extract the number to the group list
        end
      end
      
      return groupList
    end,
    
    final = function(obj, dist)
        -- provides number of very last object of a provided type.
        -- input: object type
        -- output: number
        
        local t = _M.csv.objToArray(obj..' Thru')
        if dist then return (dist + 7)
        else return tonumber(t[#t]['no.']) end
    end,
    
    info = function(object, alternate_location)
        -- filepath info
        local location = alternate_location or 'importexport'
        
        local fn = {temp = 'temp_infoinject.xml'}
        
        local dir = {}
            dir.main = gma.show.getvar('PATH')..'/'
            dir.temp = dir.main..location..'/'

        local file = {temp = dir.temp..fn.temp}
        
        
        -- function
        gma.cmd('SelectDrive 1')                                        -- select internal drive
        gma.cmd('Export '..object..' "'..fn.temp..'" /nc')                -- export object script
        io.input(file.temp)                                                -- change file input to exported script

        local script = io.read('*a')

        local index = {}
        index.infoitems = {script:find('<InfoItems>', 1)}                -- find start of InfoItems
        index.start = {script:find('<Info[^>]*>', index.infoitems[2])}    -- find start of info field within InfoItems
        index.start = index.start[2] + 1                                -- make index.start a singular number one position after <Info> tag

        index.finish = script:find('</Info>', index.start) - 1            -- find end of info field within InfoItems

        
        local info = script:sub(index.start, index.finish)                -- extract text within tags


        return info..'\n'                                                -- return extracted string
    end,
    
    rootnum_list = ST:new(),
    rootnum = function(str, objnum)
    -- CAREFUL WHAT STRING YOU SUPPLY
        local unpack = table.unpack
        str = str:lower()                                                                --set string to find to lowercase to avoid case errors
        if not _M.get.rootnum_list[str] then 
            -- tempfile info
            local temp = {}
            temp.fn = 'tempfile_getroot.csv'
            temp.dir = gma.show.getvar('PATH')..'/reports/'
            temp.file = temp.dir..temp.fn
            
            -- retrieve info from showfile
            gma.cmd('SelectDrive 1')
            gma.cmd('List /f=\"'..temp.fn..'\"')
            
            local line_locate = 0
            for line in io.lines(temp.file) do
                if line:lower():find(str) then                                                --if there is a lowercase match 
                    line_locate = tonumber(line:sub(unpack({line:find('%d+')})))            --retrieve the number following it, converted to number format
                    break
                end
            end
            
            os.remove(temp.file)        --remove temp file from system
            
            local global = false
            if line_locate and line_locate ~= 0 then
                gma.cmd('CD '..line_locate)
                gma.cmd('List /f=\"'..temp.fn..'\"')
                for line in io.lines(temp.file) do
                    if line:lower():find('global') then
                        global = true
                        break
                    end
                end
            end

            os.remove(temp.file)        --remove temp file from system
            gma.cmd('CD /')                -- return to home directory
            
            -- return result
            if line_locate == 0 then
                return false, 'string not found'
            else
                local t = ST:new()
                --if global then t = {line_locate, 1, tonumber(objnum)}
                if global then t = {line_locate, 1}
                else t = {line_locate} end
                
                _M.get.rootnum_list[str] = t
            end    
        end
        
        local t1 = _M.get.rootnum_list[str]
        local t2 = ST:new()
        for _, v in ipairs(t1) do
            t2:append(v)
        end
        t2:append(tonumber(objnum))
        gma.echo(t2:concat(', '))                            -- TROUBLESHOOT
        return t2
        
--[==[ get.rootnum test
-- require modules
local GDF = require [[GD/_GDF_v0-0-2]]

local class    = GDF.class
local gui    = GDF.gui
local dmx    = GDF.dmx
local env    = GDF.env
local data    = GDF.data
local tbl    = GDF.table
local xml    = GDF.xml
local csv    = GDF.csv
local lua    = GDF.lua
local get    = GDF.get
local set    = GDF.set
local pi    = GDF.poolitem
local plugin= GDF.plugin
local exec    = GDF.exec

local self_table = {...}



function test()
    local tprint = function(tbl)
        for i, v in ipairs(tbl) do
            gma.echo(i..'   '..v)
        end
    end
    
    local pnum = get.rootnum('plugin', 100)
    local mnum = get.rootnum('macro', 100)
    local inum = get.rootnum('image', 100)
    local lnum = get.rootnum('layout', 100)
    local enum = get.rootnum('effect', 40)
    
    tprint(pnum)
    tprint(mnum)
    tprint(inum)
    tprint(lnum)
    tprint(enum)
end



return test
--]==]
    end,
    
    
}


_M.set = {
    info = function(object, info, alternate_location)
        local location = alternate_location or 'importexport'
        
        -- filepath info
        local fn = {temp = 'temp_infoinject.xml'}
        
        local dir = {}
            dir.main = gma.show.getvar('PATH')..'/'
            dir.temp = dir.main..location..'/'

        local file = {temp = dir.temp..fn.temp}
        
        
        -- select internal drive
        gma.cmd('SelectDrive 1')
        
        -- insert marker
        gma.cmd('Unlock '..object)
        gma.cmd('Assign '..object..' /info="<SET>"')
        
        -- export object script
        gma.cmd('Export '..object..' "'..fn.temp..'" /nc')    
        
        io.input(file.temp)                    -- change input to newly-written file
        local script_src = io.read('*a')    -- import entire file to variable
        local script_output = script_src:gsub('&lt;SET&gt;', info)    --sub with actual info script
        
        -- write subbed script to original file location
        local writefile = io.open(file.temp, 'w')
        writefile:write(script_output)
        writefile:close()
        
        gma.cmd('Import "'..fn.temp..'" '..object..' /nc')    -- import over original object
        
        os.remove(file.temp)    -- remove temp file
    end,

}





---- Object-Specific Functions

_M.poolitem = { --sub-module v1.0
    generate = function(object, script, alternate_location)
        -- filepath info
        local location = alternate_location or 'importexport'
        
        local fn = {temp = 'temp_generatepoolobject.xml'}
        
        local dir = {}
            dir.main = gma.show.getvar('PATH')..'/'
            dir.temp = dir.main..location..'/'

        local file = {temp = dir.temp..fn.temp}
        
        -- generate file --
        local writefile = io.open(file.temp, 'w')
        writefile:write(script)
        writefile:close()
        
        -- import generated object --
        gma.cmd('SelectDrive 1')
        gma.cmd('Import "'..fn.temp..'" '..object)
        
        -- cleanup --
        os.remove(file.temp)
    end,
    
    checkSpace = function(poolType, start, length, exceptions)
    -- checks if range of pool spaces is empty
        -- exceptions must be in table format {}
      local finish = start + length - 1 --set our finishing point
      local emptyStatus = true
      local exceptions = exceptions or {}

      for i = start, finish do
        local obj_str = poolType..' '..tostring(i)
        if _M.get.verify(obj_str) then --if space is not empty
            gma.echo('block found')
            local exc = false                            -- exception trigger
            for _, v in ipairs(exceptions) do
                if _M.get.name(obj_str):find(v) then    -- set True if string match to exception found
                    exc = true
                    break
                end
            end
            if not exc then                             -- if not empty and no exception string found
                emptyStatus = false
                break
            end
        end
      end
      return emptyStatus
    end,
    
    advanceSpace = function(poolType, start, length, pad_before, pad_after, exceptions)
    -- padding parameters are optional - will default to 0 if not provided
        
        local pad_before    = pad_before    or 0                            --set for 0 if not provided
        local pad_after        = pad_after        or 0                            --set for 0 if not provided
        pad_before, pad_after = math.abs(pad_before), math.abs(pad_after)
        local length_actual = length + pad_before + pad_after                --length that will be used in space check
        local exceptions = exceptions or {}
        
        local finalStart
        local preroll
        if (start - pad_before) > 0 then
            preroll = pad_before
            finalStart = start - pad_before
        else
            finalStart = 1
            preroll = start - 1
        end
        
        while _M.poolitem.checkSpace(poolType, finalStart, length_actual, exceptions) == false do
            finalStart = finalStart + 1
        end
      
        finalStart = finalStart + preroll                                --offset returned answer for pre-padding
        return finalStart
    end,
}


_M.plugin = { -- plugin read/write functions v0.0.1
    -- config = {num = num, name = name, script = script, (EOL = true)}
    
    write = function(config, force_overwrite)
      local EOLnum
      if config.EOL then EOLnum = 1 
      else EOLnum = 0 end
      
      -- Establish filepath --
      local plugin = {}
      plugin.name = 'tempFile_createPlugin'
      plugin.directory = gma.show.getvar('PLUGINPATH')..'/'
      plugin.fullpathLUA = plugin.directory..plugin.name..'.lua'
      plugin.fullpathXML = plugin.directory..plugin.name..'.xml'


      -- Create text for plugin XML file --
      local xmlText = [[
<?xml version="1.0" encoding="utf-8"?>
<MA xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns="http://schemas.malighting.de/grandma2/xml/MA" xsi:schemaLocation="http://schemas.malighting.de/grandma2/xml/MA http://schemas.malighting.de/grandma2/xml/3.2.2/MA.xsd" major_vers="3" minor_vers="2" stream_vers="2">
    <Info datetime="2016-09-26T20:40:54" showfile="dummyfile" />
    <Plugin index="1" execute_on_load="]]..EOLnum..[[" name="]]..config.name..[[" luafile="]]..plugin.name..[[.lua" />
</MA>]]

      -- Write XML file to disk --
      local fileXML = assert(io.open(plugin.fullpathXML, 'w'))
      fileXML:write(xmlText)
      fileXML:close()
      
      -- Write LUA file to disk --
      local fileLUA = assert(io.open(plugin.fullpathLUA, 'w'))
      fileLUA:write(config.script)
      fileLUA:close()

      -- import plugin --
      local importcmd = 'Import \"'..plugin.name..'\" Plugin '..tostring(config.num)        -- build string for import
      if force_overwrite then importcmd = importcmd..' /o /nc' end        -- (if force-overwrite enabled)
      gma.cmd('SelectDrive 1') --select the internal drive as default read path
      gma.cmd(importcmd) --load new plugin into showfile

      -- delete temp files --
      os.remove(plugin.fullpathXML)
      os.remove(plugin.fullpathLUA)
    end,
    
    read = function(num)
        -- File Config
        local fn = {}
            fn.temp = '_plugintemp'
        local dir = {}
            dir.plugins = gma.show.getvar('PLUGINPATH')..'/'
        local file = {}
            file.xml = dir.plugins..fn.temp..'.xml'
            file.lua = dir.plugins..fn.temp..'.lua'
            
        -- export plugin
        gma.cmd('SelectDrive 1')
        gma.cmd('Export Plugin '..num..' \"'..fn.temp..'\"')
        
        local readfile = io.open(file.lua, 'r')    -- open file for reading
        local script = readfile:read('*a')        -- extract script to memory
        readfile:close()                        -- close file
    
        -- delete temp files
        os.remove(file.xml)    
        os.remove(file.lua)
        
        -- return lua script
        return script
    end
}


_M.exec = {    
    advance = function(page, exec, ct, minExec)
      -- returns table
      --\\.page .exec .str('x.xxx' format)
      
      -- advance to first available solid block of executors from provided start position
      -- if the executor number hits 100, 199, or 210, and there still are more to go, then
      -- the count restarts on the next page at the minimum executor number.
      -- if there is not enough space to pull this off on any page, the function returns nil and an error message
        
        ---- nested functions ----
        
      local function checkPass(exec, ct)
        local limits = {100, 199, 210}
        local status_pass = true
        for i = 1, #limits do
          if exec <= limits[i] and (exec+ct-1) > limits[i] then 
            status_pass = false 
            break
          end      
        end
        
        return status_pass
      end
      
      local function checkNum(num)
        local limits = {100, 199, 210}
        local status_pass = true
        for i = 1, #limits do
          if num == limits[i] then status_pass = false; break; end;
        end
        
        return status_pass
      end

      local error_msg = 'Error: minExec set too high'
      
      
      -- set minimum restart executor for a new page
      local minExec = minExec
      if not minExec then 
        local starts = {1, 101, 201}                        --base off location of provided desired executor position
        for i = 1, #starts do
            if exec > starts[i] then minExec = starts[i] end
        end
      end
      if not ct then ct = 1 end
      
      local addPage = true
      if not checkPass(minExec, ct) then addPage = false end
      
      local pageCurrent = tonumber(page)
      local execCurrent = tonumber(exec)
      local pass = false

      while pass == false do
        if checkPass(execCurrent, ct) then
            for i = 1, ct do
              if _M.exec.getstatus(pageCurrent, execCurrent+i-1) == 'EMPTY' then
                if i == ct then pass = true end
              else
                if checkNum(execCurrent+i) then
                    execCurrent = execCurrent + i
                    break
                else
                    if addPage then
                        pageCurrent = pageCurrent + 1
                        execCurrent = minExec
                        break
                    else
                        return nil, error_msg
                    end
                end
              end
            end
        else
            if addPage then
                pageCurrent = pageCurrent + 1
                execCurrent = minExec
            else
                return nil, error_msg
            end
        end
      end
      
      
      return {page = pageCurrent, exec = execCurrent, str = tostring(pageCurrent)..'.'..tostring(execCurrent)}
    end,
    
    getstatus = function(page, executor)
    -- returns 'OFF', 'ON', 'NON-SEQ', and 'EMPTY'
      local slotStatus = _M.get.handle('Executor '..page..'.'..executor); --determine if slot is empty
                
      if slotStatus then                                --return value based on executor handle
        if _M.get.handle('Executor '..page..'.'..executor..' Cue') then
            local c = _M.get.class('Executor '..page..'.'..executor..' Cue');
            if c == "CMD_SEQUENCE" then
              return 'OFF' --response for non-active executor
            elseif c == "CMD_CUE" then
              return 'ON' --response for active executor
            end
        else
            return 'NON-SEQ' --response for non-sequence executors
        end
      else
        return 'EMPTY' --response for empty executor slots
      end
    end,
    
    propstr_assign = function(str)
    -- takes the List of executor properties spit out by MA
    -- returns a string to assign those same properties to another executor
    
    -- input format example:
    ---- [[PlaybackMaster=Pb None, Width=1, SwopProtect=off]]
    -- output format example
    ---- [[/PlaybackMaster="Pb None", /width=1 /SwopProtect=off]]
        
        
        local props = {}

        -- populate props{}
        local indices = {1, 1}
        while indices[2] < #str do
          indices = {str:find('%s[^=]+[=][^,]+', indices[2])}    -- look for a pair "property=value"
          if (not indices[1]) then break end                    -- terminate loop if no comma found (end of list)
          indices[1] = indices[1] + 1                            -- reset starting point for next search

          local t = str:sub(unpack(indices))                    -- extract substring
          local key = t:sub(t:find('[^=]+'))                    -- extract portion before '=' as the key (property)

          local indices2 = {t:find('[=][^=]+')}                    -- find portion following '='
          indices2[1] = indices2[1] + 1                            -- change starting index to start AFTER '='
          local value = t:sub(unpack(indices2))                    -- extract as value

          props[key] = value                                    -- add to props list
        end

        -- turn list into string used to assign to other execs
        local str_props = ''
        for k, v in pairs(props) do
          if v:find('%s') then v = '\"'..v..'\"' end            -- use quotes around value if there is a space in it
          str_props = str_props..[[ /]]..k..[[=]]..v            -- append to final string
        end

        return str_props
    end
    
}










-- local kill_the_bug = [=[]]]=]






return _M]===]








local gma = gma
local GDF = load(GDFstr)()

local class    = GDF.class
local lo    = GDF.lo
local gui    = GDF.gui
local dmx    = GDF.dmx
local env    = GDF.env
local data    = GDF.data
local tbl    = GDF.table
local xml    = GDF.xml
local csv    = GDF.csv
local lua    = GDF.lua
local get    = GDF.get
local set    = GDF.set
local pi    = GDF.poolitem
local plugin= GDF.plugin
local exec    = GDF.exec

local self_table = {...}

local ST = class.superTable

local pBars = ST:new()
local function posgroups()


    -- User Input
    local msg = {
        pset = 'ex: 4.1, 2.8',
        seq = 'ex: 12 Thru 15 + 21; no-input or 0 = ALL',
        grp = 'ex: 101'
    }
    local input = {
        start    = gma.textinput('Starting Preset?', msg.pset),
        final    = gma.textinput('Final Preset?', msg.pset),
        grpstart= tonumber(gma.textinput('Starting Group?', msg.grp)),
        seq        = gma.textinput('Specific Sequences?', msg.seq)
    }
    
    -- terminate if missing input
    if (not input.start) or (not input.final) or (not input.grpstart) then
        goto EOF
    end
    
    pBars.space = gui.superBar:new('Finding Group Space')
    local grp_dummy = get.final('Group') + 1
    pBars.space:stop()
    gma.cmd('ClearAll')
    gma.cmd('Store Group '..grp_dummy)
    
    ---- Process Inputs
    local pset = {                                            -- parse out numbers
        type = input.start:match('(%d+)%.%d+'),
        start = input.start:match('%d+%.(%d+)'),
        final = input.final:match('%d+%.(%d+)'),
    }
    for k, v in pairs(pset) do                                -- convert to number format
        pset[k] = tonumber(pset[k])
    end

    local str_seq                                                                                -- Sequence string for search function
    if input.seq == msg.seq or tostring(input.seq) == '0' or not input.seq or input.seq == '' then str_seq = ''
    else str_seq = ' If Sequence '..input.seq end
    
    --
    pset.tbl = class.superTable:new()                                        -- table for all presets existing in range, with properties for:
    pBars.presets = gui.superBar:new('Checking Presets')
    for i = pset.start, pset.final do                                            -- number (12), full number(2.12), and name ('Flyout')
        local id = 'Preset '..pset.type..'.'..tostring(i)
        if get.verify(id) then
            pset.tbl:append{num = i, fullnum = id, name = get.label(id)}
        end
        pBars.presets:set(1, true)
    end
    pset.total = #pset.tbl
    pBars.presets:stop()
    
    local group = {start = grp_dummy + 1}    -- find starting group position
    
    gma.cmd('ClearAll')
    local text = {suffix = '', ct = ''}
    if config.grp_suffix then text.suffix = text.suffix..config.grp_suffix end
    local grp_cur = group.start
    local count = {matched = 0, unmatched = 0}
    pBars.groups = gui.superBar:new('Creating Groups')
    for i, v in ipairs(pset.tbl) do
        gma.cmd('Search /nc '..v.fullnum..str_seq)
        gma.cmd('SelFix SearchResult')
        local ct = tonumber(gma.show.getvar('selectedfixturescount'))
        if config.include_fixt_count then text.ct = '('..ct..')' end
        if ct > 0 then                                                             -- only store if fixtures present
            gma.cmd('Store /use=active Group '..grp_cur..' \"'..v.name..text.suffix..text.ct..'\"')
            grp_cur = grp_cur + 1
            count.matched = count.matched + 1
        else
            gma.cmd('Store /o /use=active Group '..grp_dummy)
            count.unmatched = count.unmatched + 1
        end 
        gma.cmd('ClearAll')
        pBars.groups:set(1, true)
        gma.sleep(0.05)
    end
    
    ---- Move Groups Toward Requested Start Point ----
    group.final = grp_cur
    group.total = group.final - group.start + 1
    group.move = pi.advanceSpace('Group', input.grpstart, group.total, 0, 0)
    gma.cmd('Move Group '..group.start..' Thru '..group.final..' At Group '..group.move)
    
    
    ---- Messages ----
    local str_seq_msg = str_seq
    if str_seq_msg == '' then str_seq_msg = '(full showfile)' end
       local message = ST:new{
        'Preset range used: '..pset.type..'.'..pset.start..' Thru '..pset.type..'.'..pset.final,
        'Sequence Filter: '..str_seq_msg,
        'Starting Group #: '..group.move,
        '',
        'Groups stored: '..count.matched,
        'Presets not used: '..count.unmatched    
    }
    message_str = message:concat('\n')
    gui.msgbox('Plugin Complete', message_str, 'Please see command line for info')
    gui.print(message_str)
    
    
    ---- Internal Cleanup ----
    gma.cmd('Delete Group '..grp_dummy)

    
    ::EOF::
end

---- CLEANUP ----
local function cleanup()
    -- Kill all progress bars
    for _, v in pairs(pBars) do
        v:stop()
    end
end

return posgroups, cleanup